/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "collutils.h"
#include "viapriv.h"

#ifdef VIADEV_RPUT_SUPPORT

/*
 * Calculates logarithm to the base 2 of a number.
*/

int log2(int np)
{
    int lgN, t;

    for (lgN = 0, t = 1; t < np; lgN++, t += t);

    return lgN;
}

/*
 * Calculates 2^x
 */
int power_two(int x)
{
    int pow = 1;

    while (x) {
        pow = pow * 2;
        x--;
    }

    return pow;
}

/* Register a block of memory */
void register_mem(void *buf, int len, VIP_MEM_HANDLE * hndl)
{
    VAPI_ret_t result;
    VAPI_mrw_t mr_in, mr_out;
    /* Register local memory */

    mr_in.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;

    mr_in.l_key = 0;
    mr_in.r_key = 0;
    mr_in.pd_hndl = viadev.ptag;
    mr_in.size = len;
    mr_in.start = (VAPI_virt_addr_t) (virt_addr_t) (buf);
    mr_in.type = VAPI_MR;

    result = VAPI_register_mr(viadev.nic, &mr_in, &(hndl->hndl), &mr_out);

    if (result != VAPI_OK) {
        vapidev_error_abort(VAPI_RETURN_ERR, "VAPI_register_mr (%s)",
                VAPI_strerror(result));
    }

    hndl->lkey = mr_out.l_key;
    hndl->rkey = mr_out.r_key;
}

/* Deregister the memory pointed to by 'hndl' */

void deregister_mem(VAPI_mr_hndl_t * hndl)
{
    VAPI_ret_t ret;

    ret = VAPI_deregister_mr(viadev.nic, *hndl);

    if (ret != VAPI_OK) {
        vapidev_error_abort(VAPI_RETURN_ERR, "VAPI_deregister_mr (%s)",
                VAPI_strerror(ret));
    }
}

void copy_recv_addr(void *addr, struct Collbuf *coll,
                    VIP_MEM_HANDLE * hndl)
{
    aint_t recv_addr;
    void *dest_addr;

    /* Cast it to integer */
    recv_addr = (aint_t) addr;

    dest_addr = (void *) ((char *) coll->l_coll->buf + COLL_FLAG_SPACE);

    /* Copy the memory handle structure to this memory
     * location */

    *((VAPI_rkey_t *) dest_addr) = hndl->rkey;

    coll->atoa->buf_hndl_addr = dest_addr;
    coll->atoa->buf_hndl_addr_len = sizeof(VAPI_rkey_t);

    dest_addr = (void *) ((char *) dest_addr + sizeof(VAPI_rkey_t));

    /* Set the value to the recv addr */
    *((aint_t *) dest_addr) = recv_addr;

    coll->atoa->rdma_recv_addr = dest_addr;
    coll->atoa->rdma_recv_addr_len = sizeof(aint_t);

    dest_addr = (void *) ((char *) dest_addr + sizeof(aint_t));
    *((char *) dest_addr) = ALLTOALL_RDMA_FLAG;
}

/* A wrap around function for performing a
 * RDMA write for my collectives
 * Contiguous RDMA write
 */
void coll_rdma_write(int dest, void *local_addr,
                     int len, VIP_MEM_HANDLE * local_hndl,
                     void *remote_addr, VIP_MEM_HANDLE * remote_hndl)
{
    VIP_MEM_HANDLE l_hndl;
    VIP_MEM_HANDLE r_hndl;

    vbuf *v = NULL;
    viadev_connection_t *c = NULL;

    /* Grab a descriptor */
    v = get_vbuf();
    assert(v != NULL);

    /* Get connection from 'dest' global rank */
    c = &viadev.connections[dest];
    assert(c != NULL);

    l_hndl.lkey = local_hndl->lkey;
    l_hndl.hndl = local_hndl->hndl;
    l_hndl.rkey = local_hndl->rkey;

    r_hndl.rkey = remote_hndl->rkey;
    r_hndl.hndl = remote_hndl->hndl;
    r_hndl.lkey = remote_hndl->lkey;

    viadev_rput(c, v, local_addr, l_hndl, remote_addr, r_hndl, len);
}

/* Wait until all send descriptors on the connection are
 * available */

void coll_all_send_complete(int dest)
{
    viadev_connection_t *c = NULL;

    c = &viadev.connections[dest];

    while(c->send_wqes_avail != viadev_sq_size) {
        MPID_DeviceCheck(MPID_NOTBLOCKING);
    }
}


void Next_Power_of_Degree(int N, int degree, int *next_pow)
{
    int high = 65536;
    int low = 1;
    while ((high > N) && (low < N)) {
        high /= degree;
        low *= degree;
    }
    if (high <= N) {
        if (high == N)          /* no defect, power of 2! */
            (*next_pow) = N;
        else
            (*next_pow) = high << 1;
    } else {                    /* condition low >= N satisfied */
        if (low == N)           /* no defect, power of 2! */
            (*next_pow) = N;
        else
            (*next_pow) = low;
    }
}


int MPID_rdma_write(void *buf,
                    int len, int src_lrank, int dest_grank, int togle,
                    /*struct MPIR_COMMUNICATOR *comm_ptr */
                    void *remote_address, VIP_MEM_HANDLE * r_handle,
                    VIP_MEM_HANDLE * l_handle)
{
    viadev_connection_t *c = &viadev.connections[dest_grank];
    VIP_MEM_HANDLE remote_handle;
    VAPI_ret_t ret4;
    VAPI_sr_desc_t sr;
    VAPI_sg_lst_entry_t sg_entry_s;


    remote_handle = *r_handle;

    sr.comp_type = VAPI_UNSIGNALED;
    sr.set_se = FALSE;
    sr.fence = FALSE;
    sr.opcode = VAPI_RDMA_WRITE;
    sr.remote_qkey = 0;
    sr.id = -1;
    sr.remote_addr = (VAPI_virt_addr_t) (MT_virt_addr_t) remote_address;
    sr.r_key = r_handle->rkey;
    sr.imm_data = -20;
    sr.sg_lst_len = 1;
    sg_entry_s.len = 1;
    sg_entry_s.lkey = l_handle->lkey;
    sg_entry_s.addr = (VAPI_virt_addr_t) (MT_virt_addr_t) buf;
    sr.sg_lst_p = &sg_entry_s;

    ret4 = VAPI_post_sr(viadev.nic, c->vi, &sr);
    if (ret4 != VAPI_OK)
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Error in posting descriptor (%s)\n",
                            VAPI_strerror(ret4));


    return MPI_SUCCESS;
}

#endif                          /* VIADEV_RPUT_SUPPORT */
