/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef _VIAPRIV_H
#define _VIAPRIV_H

#include "viaparam.h"

#include <vapi_header.h>

#include "vbuf.h"
#include "req.h" 
#include <assert.h>

#ifdef MCST_SUPPORT
#include "bcast_info.h"
#endif

#define NIC_0 0
#define SUBCHANNEL_0 0
/*
 * this is defined in src/env/debugutil.c. Should be defined in an 
 * mpir.h somewhere. 
 */

void *MPIR_Breakpoint(void);


/*
 * For setting up communication, use discriminators that include both the global
 * id and rank. If global id is unique, or almost unique, this should be good enough. 
 */

/*
 * This structure is used to implement a vbuf backlog queue
 * for posting packet sends.  If no remote credits are available
 * on a connection at the time of the packet send, it will be
 * put on the end of the queue.  As credits become available
 * the packets on the queue are sent in the order received.
 */
typedef struct _viadev_backlog_queue_t {
    int len;                   /* length of backlog queue */
    vbuf *vbuf_head;           /* head of backlog queue */
    vbuf *vbuf_tail;           /* tail of backlog queue */
} viadev_backlog_queue_t;



/* some performance counters */

#if defined(RDMA_FAST_PATH)
#define MAX_NO_COMPLETION 1

/* number of RDMA buffers for the fast path */
/* the actual number of buffers available might be NUM_RDMA_BUFFER - 1 */
#define NUM_RDMA_BUFFER 32 


#define PRIMARY_FLAG (7777777)
#define SECONDARY_FLAG (88888888)
#define VBUF_FLAG_TYPE u_int32_t 

#define NORMAL_VBUF_FLAG (222)
#define RPUT_VBUF_FLAG (333)

#define FREE_FLAG (0)
#define BUSY_FLAG (1)

#define IS_BUSY_VBUF(v) (v->padding != FREE_FLAG)

#define IS_NORMAL_VBUF(v) (v->padding == NORMAL_VBUF_FLAG)

#define IS_RPUT_VBUF(v) (v->padding == RPUT_VBUF_FLAG)

#endif

/* One of these structures per VI. Pointer to this
 * structure is also cached on the VI itself for two-way
 * lookup 
 */

typedef struct _viadev_connection_t {

    VAPI_qp_hndl_t vi[MAX_SUBCHANNELS];

    int global_rank;           
    int remote_credit[MAX_SUBCHANNELS];          /* how many vbufs I can consume on remote end. */
    int local_credit[MAX_SUBCHANNELS];           /* accumulate vbuf credit locally here */
    int preposts[MAX_SUBCHANNELS];               /* number of vbufs currently preposted */

    /* These are not necessary for reliable connections, but we use
     * them anyway as a sanity check
     * They will become necessary when we handle unreliable connections
     */
    int initialized[MAX_SUBCHANNELS];       		/* have all the initial buffers preposted */
    int send_wqes_avail[MAX_SUBCHANNELS];   		/* send Q WQES available */
    struct vbuf *ext_sendq_head[MAX_SUBCHANNELS];  	/* queue of sends which didn't fit on send Q */
    struct vbuf *ext_sendq_tail[MAX_SUBCHANNELS];  	/* queue of sends which didn't fit on send Q */


    u_int16_t next_packet_expected;   					/* for sequencing (starts at 1) */
    u_int16_t next_packet_tosend;     					/* for sequencing (starts at 1) */

    int remote_cc[MAX_SUBCHANNELS];              	/* current guess at remote sides
				 * credit count */
    /* the backlog queue for this connection. */
    viadev_backlog_queue_t backlog[MAX_SUBCHANNELS];

    /* This field is used for managing preposted receives. It is the
     * number of rendezvous packets (r3/rput) expected to arrive for
     * receives we've ACK'd but have not been completed. In general we
     * can't prepost all of these vbufs, but we do prepost extra ones
     * to allow sender to keep the pipe full. As packets come in this
     * field is decremented.  We know when to stop preposting extra
     * buffers when this number goes to zero.
     */

    int rendezvous_packets_expected; 

    /* these fields are used to remember data transfer operations
     * that are currently in progress on this connection. The 
     * send handle list is a queue of send handles representing
     * in-progress rendezvous transfers. It is processed in FIFO
     * order (because of MPI ordering rules) so there is both a head
     * and a tail. 
     *
     * The receive handle is a pointer to a single
     * in-progress eager receive. We require that an eager sender
     * send *all* packets associated with an eager receive before
     * sending any others, so when we receive the first packet of 
     * an eager series, we remember it by caching the rhandle
     * on the connection. 
     *
     */


    MPIR_SHANDLE *shandle_head; /* "queue" of send handles to process */
    MPIR_SHANDLE *shandle_tail;
    MPIR_RHANDLE *rhandle;      /* current eager receive "in progress" */

    /* these two fields are used *only* by MPID_DeviceCheck to 
     * build up a list of connections that have received new
     * flow control credit so that pending operations should be 
     * pushed. nextflow is a pointer to the next connection on the
     * list, and inflow is 1 (true) or 0 (false) to indicate whether
     * the connection is currently on the flowlist. This is needed
     * to prevent a circular list.
     */
    struct _viadev_connection_t *nextflow; 
    int inflow;

    /* used to distinguish which VIA barrier synchronozations have
     * completed on this connection.  Currently, only used during
     * process teardown.
     */
    int barrier_id;

#if defined(RDMA_FAST_PATH) || defined(MCST_SUPPORT)
	
    void *RDMA_send_buf_orig[MAX_SUBCHANNELS];
    void *RDMA_recv_buf_orig[MAX_SUBCHANNELS];

    /* RDMA buffers for sending */
    struct vbuf *RDMA_send_buf[MAX_SUBCHANNELS];

    /* RDMA buffers for receive */
    struct vbuf *RDMA_recv_buf[MAX_SUBCHANNELS];

    /* RDMA buffers needs to be registered */
    VIP_MEM_HANDLE RDMA_send_buf_hndl[MAX_SUBCHANNELS];
    VIP_MEM_HANDLE RDMA_recv_buf_hndl[MAX_SUBCHANNELS];

    /* RDMA buffer on the remote side */
    VIP_MEM_HANDLE remote_RDMA_buf_hndl[MAX_SUBCHANNELS];
    struct vbuf *remote_RDMA_buf[MAX_SUBCHANNELS];

    /* current flow control credit accumulated for remote side */
    int rdma_credit[MAX_SUBCHANNELS];

    /* pointer to the head of free send buffers */
    /* this pointer advances when packets are sent */
    int phead_RDMA_send[MAX_SUBCHANNELS];

    /* pointer to the tail of free send buffers 
     * no buffers available if head == tail */
    /* this pointer advances when we receive more credits */
    int ptail_RDMA_send[MAX_SUBCHANNELS];

    /* pointer to the head of free receive buffers 
     * this is also where we should poll for incoming
     * rdma write messages */
    /* this pointer advances when we receive packets */
    int p_RDMA_recv[MAX_SUBCHANNELS];

    /* tail of recv free buffer ??? */
    /* last free buffer at the receive side */
    int p_RDMA_recv_tail[MAX_SUBCHANNELS];

    /* we must make sure that a completion entry
     * is generated once in a while 
     * this number of be less than max outstanding WQE */
    int num_no_completion[MAX_SUBCHANNELS];

#endif
    /* counter indicating we have received remote RDMA memory
     * address and handle for this connection */
    int remote_address_received;

    /* out-of-order queue for this connection */
    vbuf *ofo_head;
    /* used for binding */
    int binding_subchannel;
    int sr_binding_subchannel;

    /* used for weighted striping */
    int s_weight[MAX_SUBCHANNELS];
    
    /* anything else ?? */
#if !defined(DISABLE_HEADER_CACHING) && defined(RDMA_FAST_PATH)
    /* cache part of the packet header to reduce eager message size */
    viadev_packet_eager_start cached_outgoing;
    viadev_packet_eager_start cached_incoming;
    int cached_hit;
    int cached_miss;
#endif
    
} viadev_connection_t;

int choose_sr_subchannel(viadev_connection_t *c);

/* function prototypes for communication scheduling,
   used when multiple channels per connection are used */

/* striping rdma messages across multiple subchannels */
void schedule_rdma(viadev_connection_t *c, char *addr, int size,
	    char** stripe_start_addr, int *stripe_size);
void schedule_rdma_even_striping(viadev_connection_t *c, char *addr, int size,
                                char** stripe_start_addr, int *stripe_size);

/* handling out-of-order messages for multiple subchannels */
int check_ofo_queues(vbuf **vbuf_addr);
void enqueue_ofo_queue(vbuf *vbuf_addr, viadev_connection_t *c);
void remove_ofo_queue(vbuf *vbuf_addr, viadev_connection_t *c);
void init_ofo_queues(void);

#if defined(RDMA_FAST_PATH)
void post_fast_rdma(viadev_connection_t *c, int len, int subchannel);
int fast_rdma_ok(viadev_connection_t *c, int data_size, int subchannel);
void release_recv_rdma(viadev_connection_t *c, vbuf *v);
void post_fast_rdma_with_completion(viadev_connection_t *c, int len, int subchannel);

#ifndef DISABLE_HEADER_CACHING
int search_header_cache(viadev_connection_t * c,
            viadev_packet_eager_start * h);
#endif

int poll_rdma_buffer(vbuf **vbuf_addr, int *out_of_order);

void viadev_incoming_rdma_address(vbuf *v, viadev_connection_t *c,
		viadev_packet_rdma_address *header);																		

#endif

#ifdef MCST_SUPPORT

typedef struct bcast_info_t {
    void *ack_buffer;
    void *ack_region;
    VIP_MEM_HANDLE ack_mem_hndl;
    void *remote_add[MAX_PROCS];
    int remote_rkey[MAX_PROCS];
    vbuf **co_buf;
    int *ack_cnt;
    int *msg_cnt;
    int ud_pkt;
    int Bcnt[MAX_PROCS];
    double **time_out;
    int *win_head;
    int *win_tail;
    int *buf_head;
    int *buf_tail;
    int *buf_credits;           //credits of co_roots
    int *buf_updates;           //freed buff count..to be sent
    int sbuf_head;              //acks at the sender side
    int sbuf_tail;
    int **recent_ack;           //latest ack for a co-root from a node
    int *min_ack;
    int *is_full;
    int ack_full;
    int init_called;
    int bcast_called;
} bcast_info_t;

#endif


typedef struct {
    VAPI_hca_hndl_t nic[MAX_NUM_HCAS];          
	/* NICs */
    VAPI_hca_port_t hca_port[MAX_SUBCHANNELS];  
	/* Port numbers */
    VAPI_pd_hndl_t ptag[MAX_NUM_HCAS];   		
	/* single protection tag for all memory registration */ 
    VAPI_qp_hndl_t *qp_hndl;    				
	/* Array of QP handles for all connections */
    VAPI_qp_prop_t *qp_prop;    				
	/* TO DO:Wil have qpnum. We cud just store qp_num instead of entire pros structure */

    VAPI_cq_hndl_t cq_hndl[MAX_NUM_HCAS];   	
	/* one cq for both send and recv */

#ifdef USE_MPD_RING
    VAPI_cq_hndl_t boot_qp_hndl[2];     /* qp's for bootstrapping */
    VAPI_qp_prop_t boot_qp_prop[2];     /* qp_prop's for bootstrapping */
    VAPI_cq_hndl_t boot_cq_hndl;        /* one cq for bootstrapping */
    int boot_tb[2][2];          /* HCA LID and QP for the ring */
#endif

    int my_hca_lid[MAX_SUBCHANNELS];  			
	/* my HCA LID */
    int addr_tb[MAX_PROCESS_NUM][MAX_SUBCHANNELS][2]; 
	/* Store HCA LID and QP num of all processes */

    unsigned long  maxtransfersize;    			
	/* max send/rdma size */

    int *lid_table;
    /* Store HCA LID of all processes */
    int *qp_table;
    /* Store HCA QP num of all processes */
    char *my_name;
    /* A string equivalent of node name */

#if defined(RDMA_FAST_PATH) && defined(SINGLE_RDMA_BUFFER)
    char *RDMA_bufs;
    VIP_MEM_HANDLE RDMA_buf_hndl;
#endif

    int np;                       				
	/* number of processes total */
    int me;                       				
	/* my process rank */
    int global_id;                				
	/* global id of this parallel app */
#ifdef VIADEV_HAVE_RDMA_LIMIT
    int outstanding_rdmas;        				
	/* number of outstanding rdma ops */
#endif

    viadev_connection_t *connections;   		
	/* array of VIs connected to other processes */
    int barrier_id;               				
	/* Used for VIA barrier operations */
#if defined(RDMA_FAST_PATH) && defined(SINGLE_RDMA_BUFFER)
    char *RDMA_bufs;
    VIP_MEM_HANDLE RDMA_buf_hndl;
#endif

#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
	viadev_packet_eager_start match_hdr;
#endif
#endif


    /* FOr UD set up */
#ifdef MCST_SUPPORT
    VAPI_qp_hndl_t ud_qp_hndl;  /*  UD QP handles */
    VAPI_qp_prop_t ud_qp_prop;
    VAPI_cq_hndl_t ud_scq_hndl;
    VAPI_cq_hndl_t ud_rcq_hndl;
    VAPI_ud_av_t av;
    VAPI_ud_av_hndl_t av_hndl;
    IB_gid_t mcg_dgid;
    IB_lid_t mclid;
    bcast_info_t bcast_info;
#endif



} viadev_info_t;

extern viadev_info_t viadev;

void viadev_init_backlog_queue(viadev_connection_t *c);
void viadev_backlog_send(viadev_connection_t *c, int subchannel);

/*
 * Function prototypes. All functions used internally by the VIA device. 
 */

void viadev_check_communication(void);

void viadev_barrier(void);

void viadev_post_send(viadev_connection_t *c, vbuf *v, int subchannel);

#ifdef VIADEV_RGET_SUPPORT
void viadev_post_rdmaread(viadev_connection_t * c, vbuf * v);
#endif

void viadev_post_rdmawrite(viadev_connection_t *c, vbuf *v, int subchannel);

void viadev_post_recv(viadev_connection_t *c, vbuf *v, int subchannel);

void viadev_ext_sendq_queue(viadev_connection_t * c, vbuf * v, int subchannel);
void viadev_ext_sendq_send(viadev_connection_t * c, int subchannel);

void viadev_copy_unexpected_handle_to_user_handle(MPIR_RHANDLE  *rhandle, 
												  MPIR_RHANDLE  *unexpected,
												  int *error_code);

void viadev_recv_r3(MPIR_RHANDLE *rhandle);
void viadev_recv_rput(MPIR_RHANDLE *rhandle);
void viadev_recv_rget(MPIR_RHANDLE *rhandle);

inline int viadev_calculate_vbufs_expected(int nbytes, 
											viadev_protocol_t protocol);

inline int viadev_eager_ok(	int len, viadev_connection_t *c, 
							int subchannel) ;

void viadev_rendezvous_reply(MPIR_RHANDLE *rhandle);

void viadev_prepost_for_rendezvous(viadev_connection_t *connection, 
								   int vbufs_expected,
								   int subchannel);
void viadev_process_recv(void *vbuf);
void viadev_process_send(void *vbuf);


void viadev_incoming_rendezvous_start(	vbuf *v, 
										viadev_connection_t *c,
								      	viadev_packet_rendezvous_start *header);

void viadev_incoming_rendezvous_reply(	vbuf *v, 
									    viadev_connection_t *c, 
								        viadev_packet_rendezvous_reply *header);

void viadev_incoming_r3_data(	vbuf *v, 
			     				viadev_connection_t *c, 
			     				viadev_packet_r3_data *header);

#ifdef VIADEV_RPUT_SUPPORT

void viadev_incoming_rput_finish(	vbuf *v, 
									viadev_connection_t *c, 
									viadev_packet_rput_finish *header);

void viadev_rput(viadev_connection_t *c, 
				 vbuf *v, 
				 void *local_address,
   		 		 VIP_MEM_HANDLE local_memhandle, void *remote_address,
		 		 VIP_MEM_HANDLE remote_memhandle, int nbytes);
void viadev_rput_finish(MPIR_SHANDLE *s);
#endif

#ifdef VIADEV_RGET_SUPPORT
void viadev_incoming_rget_finish(vbuf * v,
                                 viadev_connection_t * c,
                                 viadev_packet_rget_finish * header);
/* can leave it blank */
void viadev_rget(viadev_connection_t * c, vbuf * v, void *local_address,
                 VIP_MEM_HANDLE local_memhandle, void *remote_address,
                 VIP_MEM_HANDLE remote_memhandle, int nbytes);

void viadev_rget_finish(MPIR_RHANDLE * r);


#endif

void viadev_incoming_eager_start(vbuf *v, 
								 viadev_connection_t *c,
								 viadev_packet_eager_start *header);

void viadev_incoming_eager_next(vbuf *v, 
								viadev_connection_t *c,
				 				viadev_packet_eager_next *header);


void viadev_rendezvous_push(MPIR_SHANDLE *s);

void viadev_eager_pull(MPIR_RHANDLE *rhandle);

void viadev_send_noop_ifneeded( viadev_connection_t * c, 
								int subchannel);

void viadev_send_noop(	viadev_connection_t *c, 
						int subchannel);

char *viadev_packet_to_string(VAPI_qp_hndl_t vi, vbuf *v);

extern char nullsbuffer, nullrbuffer;

#if defined(VIADEV_RPUT_SUPPORT)||defined(VIADEV_RGET_SUPPORT)
void viadev_register_recvbuf_if_possible(MPIR_RHANDLE *rhandle);
#endif

void viadev_unimplemented(char *s);

int vi_to_grank(VAPI_qp_hndl_t vi);
void my_memcopy(char *dest, char *src, int nbytes);


/* 
 * on 32 bit machines, we identify (for now) a handle by its address, using the 
 * address for immediate data. 
 * on 64 bit machines, we need to use the pointer index stuff since we don't
 * have 64 bits of immediate data. 
 * We will also need the index stuff when we implement reliability. 
 */

#define REQ_TO_ID(h) ((request_id_t)(h))
#define ID_TO_REQ(id)((union MPIR_HANDLE *)(id))


#define PREPOST_VBUF_RECV(c, subchannel)  {              			\
		vbuf *v = get_vbuf();                    					\
			vbuf_init_recv(v, VBUF_BUFFER_SIZE, subchannel); 		\
			viadev_post_recv(c, v, subchannel);              		\
			c->local_credit[subchannel]++;                      	\
			c->preposts[subchannel]++;                          	\
}

#ifdef MCST_SUPPORT

#define PREPOST_UD_VBUF_RECV() {                                    \
        vbuf *v = get_vbuf();                                       \
        vbuf_init_recv(v, VBUF_BUFFER_SIZE,SUBCHANNEL_0);           \
        viadev_post_ud_recv(v);                                     \
}

#endif

/*
 * this should be a real function. Need to think carefully about
 * what has to be done.
 * note model does MPID_UnpackMessageComplete, etc.
 */

#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)

#include "dreg.h" 
#include <unistd.h>
#define SEND_COMPLETE(s) {                                          \
    s->is_complete = 1;                                             \
    if (s->dreg_entry != NULL) {                                    \
        dreg_decr_refcount(s->dreg_entry);                          \
    }                                                               \
    if (s->finish != NULL) {                                        \
        s->finish(s);                                               \
    }																\
	s->dreg_entry = NULL;                                           \
    if (s->ref_count == 0) {                                        \
        switch (s->handle_type) {                                   \
            case MPIR_SEND:                                         \
            {                                                       \
                MPID_SendFree(s);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_SEND:                              \
            {                                                       \
                MPID_PSendFree(s);                                  \
                break;                                              \
            }                                                       \
            default:                                                \
                vapidev_error_abort(GEN_EXIT_ERR, "SEND_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#define RECV_COMPLETE(r) {                                          \
    r->is_complete = 1;                                             \
    r->s.MPI_ERROR = MPI_SUCCESS;                                   \
    r->s.count = r->bytes_as_contig;                                \
    if (r->dreg_entry != NULL) {                                    \
        dreg_decr_refcount(r->dreg_entry);                          \
    }                                                               \
    if (r->finish != NULL) {                                        \
        r->finish(r);                                               \
    }       														\
	r->dreg_entry = NULL;                                           \
    if (r->ref_count == 0) {                                        \
        switch (r->handle_type) {                                   \
            case MPIR_RECV:                                         \
            {                                                       \
                MPID_RecvFree(r);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_RECV:                              \
            {                                                       \
                MPID_PRecvFree(r);                                  \
                break;                                              \
            }                                                       \
            default:                                                \
                vapidev_error_abort(GEN_EXIT_ERR, "RECV_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#else

#define SEND_COMPLETE(s) {                                          \
    s->is_complete = 1;                                             \
    if (s->finish != NULL) s->finish(s);                            \
    if (s->ref_count == 0) {                                        \
        switch (s->handle_type) {                                   \
            case MPIR_SEND:                                         \
            {                                                       \
                MPID_SendFree(s);                                   \
                    break;                                          \
            }                                                       \
            case MPIR_PERSISTENT_SEND:                              \
            {                                                       \
                MPID_PSendFree(s);                                  \
                break;                                              \
            default:                                                \
                vapidev_error_abort(GEN_EXIT_ERR, "SEND_COMPLETE invalid type\n");    \
            }                                                       \
        }                                                           \
    }                                                               \
}

#define RECV_COMPLETE(r) {                                          \
    r->is_complete = 1;                                             \
    r->s.MPI_ERROR = MPI_SUCCESS;                                   \
    r->s.count = r->bytes_as_contig;                                \
    if (r->finish != NULL) {                                        \
        r->finish(r);                                               \
    }                                                               \
    if (r->ref_count == 0) {                                        \
        switch (r->handle_type) {                                   \
            case MPIR_RECV:                                         \
            {                                                       \
                MPID_RecvFree(r);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_RECV:                              \
            {                                                       \
                MPID_PRecvFree(r);                                  \
                    break;                                          \
            }                                                       \
            default:                                                \
                vapidev_error_abort(GEN_EXIT_ERR, "RECV_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#endif              /* RPUT, RGET */




#define MPID_CommInit(oldcomm,newcomm) MPI_SUCCESS
#define MPID_CommFree(oldcomm) MPI_SUCCESS

#endif /* _VIAPRIV_H */
